(****************************************************************************)
(* An implementation of Higher-Order Pattern Unification in SML             *)
(* Copyright (C) 2004 Gopalan Nadathur and Natalie Linnell                  *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(*
   This file contains code for head normalizing terms based on a lazy, 
   graph reduction approach. The main components of the code are 
 
         NORM      a signature for a normalization module
         NormFunc  a functor that produces a normalization structure
                   when given a realization of terms

   The code is dependent on an implementation of terms like that found in 
   terms.sml to be available in the context of use. In particular, the
   signature and functor defined here can only be understood in a situation
   where the TERM signature from terms.sml is already known. 

   The normalization routine is based on ideas in the paper 

       Nadathur, G. and Qi, X. Explicit Substitutions in the Reduction
       of Lambda Terms, Proceedings of the Fifth ACM-SIGPLAN
       International Conference on Principles and Practice of
       Declarative Programming, pages 195-206, 2003.

   a version of which can be found from the URL

       http://www-users.cs.umn.edu/~gopalan/recentpapers.html.
*)


signature NORM =
sig

structure Term : TERM

val hnorm : Term.term -> Term.term

end



functor NormFunc(structure Term : TERM) : NORM =
struct

structure Term = Term

fun make_explicit(t,0,0,nil) = t
  | make_explicit(ref(Term.lam(n,t)),ol,nl,e) = 
      ref(Term.lam(n,ref(Term.susp(t,ol+n,nl+n,(Term.add_dummies e n nl)))))

fun hn_lazy(term as ref(Term.const _),ol,nl,e,whnf) = (term,0,0,nil)
  | hn_lazy(term as ref(Term.var _),ol,nl,e,whnf) = (term,0,0,nil)
  | hn_lazy(term as ref(Term.dB _),0,0,nil,whnf) = (term,0,0,nil)
  | hn_lazy(term as ref(Term.dB i),ol,nl,e,whnf) =
      if (i > ol) then (ref(Term.dB(i-ol+nl)),0,0,nil)
      else (fn (Term.dum(l)) => (ref(Term.dB(nl-l)),0,0,nil)
             | (Term.bndg(t,l)) => 
                 (fn t as ref(Term.susp(t2,ol',nl',e')) => 
                       let val s = hn_lazy(t2,ol',nl'+nl-l,e',whnf)
                       in (if (nl = l)
                           then (Term.bind t (make_explicit s))
                           else ()); 
                          s
                       end
                   | t => hn_lazy(t,0,nl-l,nil,whnf)
                 ) (Term.deref t)) (Term.nth(e, i))
  | hn_lazy(term as ref(Term.lam _),ol,nl,e,true) = (term,ol,nl,e) 
  | hn_lazy(term as ref(Term.lam (n,t)),ol,nl,e,false) =
      let val (t',_,_,_) = 
            if ((ol=0) andalso (nl=0)) then hn_lazy((Term.deref t),0,0,nil,false) 
            else hn_lazy((Term.deref t),ol+n,nl+n,(Term.add_dummies e n nl),false)
          fun form_abst (ref (Term.lam (n',s))) = ref (Term.lam (n+n',s))
            | form_abst s = ref (Term.lam (n,s))
      in ((form_abst t'),0,0,nil) 
      end  
  | hn_lazy(term as ref(Term.app(t1,args)),ol,nl,e,whnf) =
      let 
        val argwrap = if ((ol = 0) andalso (nl=0))
                      then (fn x => x)
                      else (fn x => ref(Term.susp(x,ol,nl,e)))

        fun contract_and_norm (ref(Term.lam(n,t))) args fol fnl fe nargs _ =
              let 
                fun extend_env 0 args e = (args,e)
                  | extend_env n nil e = (nil,e)
                  | extend_env n (a::rargs) e =
                      extend_env (n-1) rargs ((Term.bndg((argwrap a),fnl))::e)
                val (rargs,nenv) = (extend_env n args fe)
              in if (n >= nargs)
                 then hn_lazy((Term.makelambda (n-nargs) t),
                                        fol+nargs,fnl,nenv,whnf)
                 else let val (t',ol',nl',e') = hn_lazy(t,fol+n,fnl,nenv,true)
                      in (contract_and_norm t' rargs ol' nl' e' (nargs - n) true)
                      end
              end
          | contract_and_norm (ref (Term.app(t,args'))) args _ _ _ _ _ =
             if ((ol=0) andalso (nl=0))
             then (ref(Term.app(t,args' @ args)),0,0,nil)
             else (ref(Term.app(t,
                                args' @ (map (fn a => 
                                           ref(Term.susp(a,ol,nl,e))) args))),
                   0,0,nil)
          | contract_and_norm t args _ _ _ _ changed =
             if ((ol=0) andalso (nl=0))
             then if changed then (ref(Term.app(t,args)),0,0,nil)
                  else (term,0,0,nil)
             else (ref(Term.app(t,map (fn a => 
                                         ref(Term.susp(a,ol,nl,e))) args)),
                   0,0,nil)

        val (f,fol,fnl,fe) = hn_lazy((Term.deref t1),ol,nl,e,true)

        val s = (contract_and_norm f args fol fnl fe (length args) false)
      
      in if ((ol=0) andalso (nl=0))
         then let val (t',ol',nl',e') = s
              in (if ((ol'=0) andalso (nl'=0))
                  then (Term.bind term t')
                  else ()); s
              end
         else s
      end
  | hn_lazy(term as ref(Term.susp(t,ol,nl,e)),ol',nl',e',whnf) =
      let val s =  hn_lazy(t,ol,nl,e,whnf)
      in (Term.bind term (make_explicit s));
         if ((ol'=0) andalso (nl'=0)) then s
         else hn_lazy(term,ol',nl',e',whnf)
      end


(* Head normalization function. Reduction is done destructively and the 
   head normalized result is returned in dereferenced form. Also note that 
   iterated abstractions and applications are collected at the top level.
   The function expects a dereferenced term as input.
*)
fun hnorm term = 
      let val (hnterm,_,_,_) = hn_lazy((Term.deref term),0,0,nil,false) 
      in hnterm 
      end

end
